/*
   getcat_buildq_BSC.c

  Purpose:

   Build the query string to read BSC "ROSS-custom" MySQL DB "RossCats.BSC".

   It only requires the RA, DEC corners (in deg.) of a rectangle,
   or the center of a square ("RAdeg DECdeg" or "hh mm ss dd mm ss") and
   its side in arcmin.
   The resulting MySQL query returns all the data for the objects found in that
   rectangle/square. No sorting is performed.

   LN@IASF-CNR, June 2003                   ( Last change: 18/11/2004 )

  18/11/2004: Allowed to have ':' in coords.

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

extern double deg_ra(char *ra_str);
extern double deg_dec(char *dec_str);

// Number of required arguments (could be 3)
#define NVALS   4

#define LSTR1   27
#define LRAN    4
#define LDECN   5
#define LRAV    9
#define LDECV   9
#define LAND    5
#define L2      2

/* degrees to radians */
#define DEG2RAD 1.745329251994329577E-2

// Largest area for selection (deg^2)
#define MAXAREA_CAT_BSC  600

char *
getcat_buildq_BSC(int ac,char *av[]) {
  int args = 0;                 // Start to parse argument no 1
  unsigned short is_neg;        // Check for negative Dec (like -00 ...)
  unsigned int nct=0;           // Total No of characters for the query string 
  unsigned int is_square=0;     // Toggle for input square center and size
  unsigned int ra0border=0;     // Toggle for negative start RA range
  double ra[NVALS], dec[NVALS], side_arcm=0., hside_deg=0.;
  double rac=0., decc=0., ra0, ra2, dec0, dec2;
  float ramin,ramax, decmin,decmax;

  char sel_str1[] = {
       "SELECT * FROM BSC WHERE(("};
  char rad_strn[] = {
       "RAd"};
  char decd_strn[] = {
       "DECd"};
  char sel_str2[] = {
       "))"};
// do not order       ") ORDER BY RAcs ASC"};
  char rad_strv[LRAV], decd_strv[LDECV], dec_deg[4];

  char *comm_str;

//  ac--;
/* There must be at least 3 args. */
  if (ac != NVALS)
  {
    if (ac == 7)  // RA (hh mm ss) Dec. (dd mm ss) Rad
    {
      rac = (atof(av[args++])+atof(av[args++])/60.+atof(av[args++])/3600.)*15.;
/* Cannot use strncmp as the string could start with blanks */
      strcpy(dec_deg,av[args++]);
      if ( strchr(dec_deg,'-') != NULL )
      {
        is_neg = 1;
        decc = - atof(dec_deg);
      } else {
        is_neg = 0;
        decc = atof(dec_deg);
      }
      decc = decc+atof(av[args++])/60.+atof(av[args++])/3600.;
      if ( is_neg ) decc *= -1;
    } else if (ac != 3)
    {
      return("\0");
    } else {
/* Check if RA and Dec are given as string with ":" */
      if (strchr(av[args], ':'))
      {
        rac  = deg_ra(av[args++]);
        decc = deg_dec(av[args++]);
      } else {
        rac  = atof(av[args++]);
        decc = atof(av[args++]);
      }
    }
    is_square = 1;
  }

  if (is_square)
  {
    side_arcm = atof(av[args++]);
//printf("rac,decc,side_arcm %f %f %f\n",rac,decc,side_arcm);
    if (side_arcm <= 0.)
    {
      printf("You gave a negative side: %f\n\n",side_arcm);
      return("\0");
    }
    hside_deg = side_arcm/2./60.;
     ra[0] =  rac - hside_deg/cos(decc*DEG2RAD);
    dec[0] = decc - hside_deg;
     ra[2] =  rac + hside_deg/cos(decc*DEG2RAD);
    dec[2] = decc + hside_deg;
  } else {
     ra[0] = atof(av[args++]);
    dec[0] = atof(av[args++]);
     ra[2] = atof(av[args++]);
    dec[2] = atof(av[args++]);
    rac  = (ra[0]+ra[2])/2.;
    decc = (dec[0]+dec[2])/2.;
  }
/* Order ranges */
    ra0 = (ra[0] < ra[2]) ? ra[0] : ra[0] - 360.;
    ra2 = (ra[2] > ra0) ? ra[2] : ra0;
    dec0 = (dec[0] < dec[2]) ? dec[0] : dec[2];
    dec2 = (dec[2] > dec[0]) ? dec[2] : dec[0];

  if (ra0 == ra2 || dec0 == dec2)
  {
    printf("%s: cannot give identical RA or DEC\n",av[0]);
    return("\0");
  }

  if ((ra2-ra0)*cos(decc*DEG2RAD) * (dec2-dec0) > MAXAREA_CAT_BSC)
  {
    printf("Region to large. Max %d deg^2\n\n",MAXAREA_CAT_BSC);
    return("\0");
  }


/* Check for border effects: not all cases are considered here! */
    if (ra2 <= 0.)
    {
      printf("Negative RA: %f %f\n\n",ra0,ra2);
      return("\0");
    }
    if (ra0 < 0.)
    {
      ra0 += 360.;
      ra0border = 1;
    }

    if (dec0 >= 90. || dec2 <= -90.)
    {
      printf("DEC out of range: %f %f\n\n",dec0,dec2);
      return("\0");
    }
    dec0 = (dec0 < -90.) ? -90. : dec0;
    dec2 = (dec2 >  90.) ?  90. : dec2;

   ra[0] = ra0;
   ra[2] = ra2;
  dec[0] = dec0;
  dec[2] = dec2;
   ra[1] = ra[0];
  dec[1] = dec[2];
   ra[3] = ra[2];
  dec[3] = dec[0];

/* Define integer ranges */
  ramin = ra[0];
  ramax = ra[2];
  decmin = dec[0];
  decmax = dec[2];

/* Build the coordinates string */
  nct += 2*(LAND+LRAN+1+LRAV) + 2*(LAND+LDECN+1+LDECV) + 2*L2;
  nct += LSTR1 + 2*L2;

  comm_str = (char *) malloc(nct);

//printf("No. char. allocated for query string: %d \n",nct);

/* Build the query string */
  strcpy(comm_str,sel_str1);

  strcat(comm_str,rad_strn);
  strcat(comm_str,">");
  sprintf(rad_strv,"%8.4f",ramin);
  strcat(comm_str,rad_strv);

  if ( ra0border )
  {
    strcat(comm_str," OR ");
  } else {
    strcat(comm_str," AND ");
  }

  strcat(comm_str,rad_strn);
  strcat(comm_str,"<");
  sprintf(rad_strv,"%8.4f",ramax);
  strcat(comm_str,rad_strv);

  strcat(comm_str,") AND (");

  strcat(comm_str,decd_strn);
  strcat(comm_str,">");
  sprintf(decd_strv,"%8.4f",decmin);
  strcat(comm_str,decd_strv);
  strcat(comm_str," AND ");
  strcat(comm_str,decd_strn);
  strcat(comm_str,"<");
  sprintf(decd_strv,"%8.4f",decmax);
  strcat(comm_str,decd_strv);

  strcat(comm_str,sel_str2);

//printf("%d\n",strlen(comm_str));
  return(comm_str);

}
