/*
   gsc23tdb

  Purpose:

   Read GSC2.3 (v1) FITS files and writes them into a "Custom" MySQL DB
   (see "MyServer_def.h").

   Simple version for the IMPRS DB course.

SQL to create table:

USE MyCats;
CREATE TABLE GSC23 (
gscID2     MEDIUMINT UNSIGNED NOT NULL default 0,
RAcs       INT UNSIGNED NOT NULL default 0,
DECcs      INT NOT NULL default 0,
Epoch00c   SMALLINT NOT NULL default 0,
RaPMdm     SMALLINT NOT NULL default 0,
DecPMdm    SMALLINT NOT NULL default 0,
RaPMerrdm  TINYINT UNSIGNED NOT NULL default 0,
DecPMerrdm TINYINT UNSIGNED NOT NULL default 0,
Fmm        SMALLINT NOT NULL default 0,
Jmm        SMALLINT NOT NULL default 0,
Vmm        SMALLINT NOT NULL default 0,
Nmm        SMALLINT NOT NULL default 0,
J_Fcm      TINYINT NOT NULL default 0,
F_Ncm      TINYINT NOT NULL default 0,
Class      TINYINT UNSIGNED NOT NULL default 0,
EXTcs      SMALLINT UNSIGNED NOT NULL default 0,
Eccen      TINYINT UNSIGNED NOT NULL default 0,
PosAng     TINYINT UNSIGNED NOT NULL default 0,
sourceStat INT UNSIGNED NOT NULL default 0)
 ENGINE=MyISAM MAX_ROWS=1200000000;


  LN@INAF-IASF, December 2014                   ( Last change: 03/12/2014 )
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include "my_db.h"
#include "fitsio.h"
#include "MyServer.h"
#include "ConvConst.h"

/* Number of columns in the original GSC2.3 catalogue */
#define NNCOLS 50

/* Reference epoch to subtract to meanEpoch: Tycho stars have ref. < 1950. */
#define REF_EPOCH 1900.0

/* non-existing data markers: for non negative fields it is positive */
#define NODATAm9999  -9999
#define NODATAp9999  +9999
#define NODATAm32768 -32768
#define NODATAp255   +255
#define NODATAm128   -128
#define NODATA0       0


// Version ID string
#define VERID "Ver 1.1, 13-10-2004, LN@INAF-IASF"


int
main(int argc, char *argv[])
{
  fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */
  static int hdunum=3;  /* HDU of interest */
  int status, hdutype,  nfound, anynull, i, ret, lt, lroot;
  char ins_parvals[220], ins_comm[400];

/* What to read from the files: memory allocated below */
  double *RArad=NULL, *Decrad=NULL;
  float *meanEpoch=NULL, *RaPM=NULL, *DecPM=NULL, *RaPMerr=NULL, *DecPMerr=NULL,
        *FpgMag=NULL, *JpgMag=NULL, *VMag=NULL, *NpgMag=NULL,
        *semiMajorAxis=NULL, *eccentricity=NULL, *PosAngle=NULL;
  long *gscID2=NULL, *Class=NULL, *sourceStatus=NULL;

/* What to write into the DB: Epoch00c not declared "unsigned" because of
   possible values < 1900.0 (was 1950.0) */
  unsigned long RAcs;
  long DECcs;
  short Epoch00c,RaPMdmas,DecPMdmas,RaPMerr2mas,DecPMerr2mas,Fmm,Jmm,Vmm,Nmm,Ecc,PosAng;
  short JminusFdm, FminusNdm;
/* Class is encoded in bit position */
  unsigned short Cl, EXTcs;
  long gscID;

/* Columns indices of relevant fields: these are different for GSC2.2 and
   GSC2.3 version 0, 1 (see GSC23v1_entry_desc.txt in the aux dir.)

TTYPE1 = 'gscID2
TTYPE4 = 'RightAsc'
TTYPE5 = 'Declination'
TTYPE6 = 'PositionEpoch'
TTYPE9 = 'raProperMotion'
TTYPE10 = 'decProperMotion'
TTYPE11 = 'raProperMotionErr'
TTYPE12 = 'decProperMotionErr'
TTYPE14 = 'FpgMag
TTYPE17 = 'JpgMag
TTYPE20 = 'VMag
TTYPE23 = 'NpgMag
TTYPE47 = 'classification'
TTYPE48 = 'semiMajorAxis'
TTYPE49 = 'eccentricity'
TTYPE50 = 'positionangle'
*/

  const int colget[]={1, 4, 5, 6, 9, 10, 11, 12, 14, 17, 20, 23,
                      47, 48, 49, 50};
  const int sourceStatus_col=51;
//  const int bad_flag=1111111118;

  long frow, felem, nelem, cur_nelem=0, longnull, naxes[2];
  float floatnull;

  char db_tabname[40], username[16], password[16];
  char filename[132], filelis[80], fileroot[9];
  const char filedblent[]={"dblent_gsc23v1.lis"};

  FILE *inlis, *dblent; // *oldepoch;


  if (argc < 2)
  {
    printf("%s  %s\n\n",argv[0], VERID);
    printf("Usage:\n"
           "       %s GSC2_list_file_name [ DB_table_name ]\n",argv[0]);
    exit(1);
  }
  strcpy(filelis,*++argv);

  if (argc > 2)
    strcpy(db_tabname,*++argv);
  else
    strcpy(db_tabname,"GSC23");


/* Connect to the DB */

  ret=db_init(0);
  if (!ret)
  {
    printf("Can't set CONNECT_TIMEOUT for MySQL connection.");
    return(1);
  }

  printf("Enter username and password: ");
  scanf("%s %s", username, password);


/* db_connect(0, hostname, username, password, dbname), returns 0 on error */
//  ret=db_connect(0, "localhost", "ross", "rosspasswd", "RossCats");
  ret=db_connect(0, "localhost", username, password, SERVER_DB);

  if (!ret) { printf("DB error: %s\n", db_error(0));  return(1); }

/* Open the double entry (diff <~ .002") list file */
  if ( (dblent = fopen(filedblent,"a")) == NULL )
  {
    printf("Could not open double entry list file %s \n",filedblent);
    exit(3);
  }

/*
  if ( (oldepoch = fopen(fileoldepoch,"a")) == NULL )
  {
    printf("Could not open old epochs list file %s \n",fileoldepoch);
    exit(4);
  }
*/


/* Loop on list of files */

  if ( (inlis = fopen(filelis,"r")) == NULL )
  {
    printf("Could not open regions list file %s \n",filelis);
    exit(2);
  }

  while( fgets(filename,131,inlis) != NULL )
  {

    lt = strlen(filename);
    filename[lt-1] = '\0';    // remove the "\n"

    if ( strstr(filename,".gz") != NULL ) lroot = 16; else lroot = 13;
/*
    this_ext = strstr(filename,dot);
    le = strlen(this_ext);
    lf = lt - le;
*/

    strncpy(fileroot,&filename[lt-lroot],8);
    fileroot[8] = '\0';    // terminate string

/* Check for lower case filenames */
    if ( fileroot[0] == 's' ) fileroot[0] = 'S';
    else if ( fileroot[0] == 'n' ) fileroot[0] = 'N';

/* Get ID */
    printf("Now processing %s:  ", filename);

    status = 0;

    if ( fits_open_file(&fptr, filename, READONLY, &status) )
         fits_report_error(stderr, status);

/* move to the HDU */
    if ( fits_movabs_hdu(fptr, hdunum, &hdutype, &status) )
         fits_report_error(stderr, status);

    if (hdutype == ASCII_TBL)
    {
      printf("\nThere is an ASCII table in HDU %d:\n", hdunum);
      exit(1);
    }
/* Read the number of rows */
    if ( fits_read_keys_lng(fptr, "NAXIS", 1, 2, naxes, &nfound, &status) )
       fits_report_error(stderr, status);

    frow      = 1;
    felem     = 1;
    nelem     = naxes[1];
    longnull  = 0;
    floatnull = 0.;

    printf("%ld objects\n",nelem);

/* Re-Allocate memory for the selected fields */
    if (cur_nelem < nelem)
    {
      cur_nelem = nelem;
      if (!(gscID2 = (long*) realloc(gscID2,sizeof(long)*nelem)))
        exit (-1);
      if (!(RArad = (double*) realloc(RArad,sizeof(double)*nelem)))
        exit (-1);
      if (!(Decrad = (double*)realloc(Decrad,sizeof(double)*nelem)))
        exit (-1);
      if (!(meanEpoch = (float*)realloc(meanEpoch,sizeof(float)*nelem)))
        exit (-1);
      if (!(RaPM     = (float*)realloc(RaPM,sizeof(float)*nelem)))
        exit (-1);
      if (!(DecPM    = (float*)realloc(DecPM,sizeof(float)*nelem)))
        exit (-1);
      if (!(RaPMerr  = (float*)realloc(RaPMerr,sizeof(float)*nelem)))
        exit (-1);
      if (!(DecPMerr = (float*)realloc(DecPMerr,sizeof(float)*nelem)))
        exit (-1);
      if (!(FpgMag = (float*)realloc(FpgMag,sizeof(float)*nelem)))
        exit (-1);
      if (!(JpgMag = (float*)realloc(JpgMag,sizeof(float)*nelem)))
        exit (-1);
      if (!(VMag = (float*)realloc(VMag,sizeof(float)*nelem)))
        exit (-1);
      if (!(NpgMag = (float*)realloc(NpgMag,sizeof(float)*nelem)))
        exit (-1);
      if (!(Class = (long*)realloc(Class,sizeof(long)*nelem)))
        exit (-1);
      if (!(semiMajorAxis = (float*)realloc(semiMajorAxis,sizeof(float)*nelem)))
        exit (-1);
      if (!(eccentricity = (float*)realloc(eccentricity,sizeof(float)*nelem)))
        exit (-1);
      if (!(PosAngle = (float*)realloc(PosAngle,sizeof(float)*nelem)))
        exit (-1);
      if (!(sourceStatus = (long*)realloc(sourceStatus,sizeof(long)*nelem)))
        exit (-1);
    }

/* Read columns */
    if ( fits_read_col(fptr, TLONG, colget[0], frow, felem, nelem, &longnull,  gscID2,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TDOUBLE, colget[1], frow, felem, nelem, &longnull,  RArad,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TDOUBLE, colget[2], frow, felem, nelem, &longnull,  Decrad,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[3], frow, felem, nelem, &floatnull, meanEpoch,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[4], frow, felem, nelem, &floatnull, RaPM,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[5], frow, felem, nelem, &floatnull, DecPM,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[6], frow, felem, nelem, &floatnull, RaPMerr,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[7], frow, felem, nelem, &floatnull, DecPMerr,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[8], frow, felem, nelem, &floatnull, FpgMag,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[9], frow, felem, nelem, &floatnull, JpgMag,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[10], frow, felem, nelem, &floatnull, VMag,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[11], frow, felem, nelem, &floatnull, NpgMag,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TLONG, colget[12], frow, felem, nelem, &longnull,  Class,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[13], frow, felem, nelem, &floatnull, semiMajorAxis,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[14], frow, felem, nelem, &floatnull, eccentricity,
                  &anynull, &status) )
    fits_report_error(stderr, status);
    if ( fits_read_col(fptr, TFLOAT, colget[15], frow, felem, nelem, &floatnull, PosAngle,
                  &anynull, &status) )
    fits_report_error(stderr, status);

/* This is for check purposes */
    if ( fits_read_col(fptr, TLONG, sourceStatus_col, frow, felem, nelem, &floatnull, sourceStatus,
                  &anynull, &status) )
    fits_report_error(stderr, status);

    if ( fits_close_file(fptr, &status) )
         fits_report_error(stderr, status);

/* First lock the table */
    strcpy(ins_comm, "LOCK TABLES ");
    strcat(ins_comm, db_tabname);
    strcat(ins_comm, " WRITE");
    ret=db_query(0, ins_comm);
    if (!ret) { printf("DB error: %s\n", db_error(0));  return(1); }

/* Write to DB checking for NULL fields */
    for (i=0;i<nelem;i++)
    {
 
      if (isnan(gscID2[i])) gscID2[i] = NODATA0;
/*
   NULL PM values give marked values also for the errors and vice-versa
   Also values less then 1 are reset so that the converted minimum value
   for RaPMerr2mas and DecPMerr2mas is 1 (see below).
*/
      if (isnan(RaPM[i]))
      {
        RaPM[i] = NODATA0;
        RaPMerr[i] = NODATA0;
      } else {
        if (isnan(RaPMerr[i]))
        {
          RaPM[i] = NODATA0;
          RaPMerr[i] = NODATA0;
        } else {
          if (RaPMerr[i] <= 1.) RaPMerr[i] = 1.5;
        }
      }

      if (isnan(DecPM[i]))
      {
        DecPM[i] = NODATA0;
        DecPMerr[i] = NODATA0;
      } else {
        if (isnan(DecPMerr[i]))
        {
          DecPM[i] = NODATA0;
          DecPMerr[i] = NODATA0;
        } else {
          if (DecPMerr[i] <= 1.) DecPMerr[i] = 1.5;
        }
      }

      if (isnan(FpgMag[i])) FpgMag[i] = NODATAm32768;
      if (isnan(JpgMag[i])) JpgMag[i] = NODATAm32768;
      if (isnan(VMag[i])) VMag[i] = NODATAm32768;
      if (isnan(NpgMag[i])) NpgMag[i] = NODATAm32768;

      if (isnan(meanEpoch[i])) meanEpoch[i] = NODATAm9999;
//      if (isnan(Class[i])) Class[i] = 4;  // mark "unclassified" those undefined
      if (isnan(Class[i])) Class[i] = NODATAp255;
      if (isnan(eccentricity[i])) eccentricity[i] = NODATAp255;

/* This is peculiar to GSC2.2: use 99.9 for Ecc. semiM. and PosA. */
      if (eccentricity[i] > 1.0)
      {
        semiMajorAxis[i] = NODATA0;
        eccentricity[i] = NODATAp255;
        PosAngle[i] = NODATAp255;
      } else {
        if (isnan(semiMajorAxis[i])) semiMajorAxis[i] = NODATA0; // not negative
        if (isnan(PosAngle[i])) PosAngle[i] = NODATAp255;
      }

/* To centi-arcsec etc. */
      gscID   = (long)(gscID2[i]);
      RAcs    = (unsigned long)(RArad[i]*RAD2ARCSEC*100);
      DECcs   = (long)(Decrad[i]*RAD2ARCSEC*100);
      Epoch00c = (short)(meanEpoch[i]==NODATAm9999 ? meanEpoch[i] :
               ((meanEpoch[i]-REF_EPOCH)*100 + 0.5));
      RaPMdmas  = (short)(RaPM[i]*10);
      DecPMdmas = (short)(DecPM[i]*10);
      RaPMerr2mas = (short)(RaPMerr[i]/2. + 0.5);
      DecPMerr2mas= (short)(DecPMerr[i]/2. + 0.5);
      Fmm     = (short)(FpgMag[i]==NODATAm32768 ? FpgMag[i] : FpgMag[i]*1000);
      Jmm     = (short)(JpgMag[i]==NODATAm32768 ? JpgMag[i] : JpgMag[i]*1000);
      Vmm     = (short)(VMag[i]==NODATAm32768 ? VMag[i] :  VMag[i]*1000);
      Nmm     = (short)(NpgMag[i]==NODATAm32768 ? NpgMag[i] : NpgMag[i]*1000);
//      Cl      = (unsigned short)(1<<Class[i]);
      Cl      = (unsigned short)(Class[i]);
      EXTcs   = (unsigned short)(semiMajorAxis[i]*100);
//      Ecc     = (short)(eccentricity[i]==NODATAm9999 ? eccentricity[i] :
      Ecc     = (unsigned short)(eccentricity[i]==NODATAp255 ? eccentricity[i] :
               eccentricity[i]*100);
//      PosAng  = (short)PosAngle[i];
      PosAng  = (unsigned short)(PosAngle[i]);

/* This should never happen, still there were such cases */
      if (RAcs > 129599999) RAcs -= 129600000;

      if (Jmm != NODATAm32768 && Fmm != NODATAm32768)
        { JminusFdm = (short)((JpgMag[i] - FpgMag[i]) * 10); }
      else
        { JminusFdm = NODATAm128; }

      if (Jmm != NODATAm32768 && Fmm != NODATAm32768)
        { FminusNdm = (short)((FpgMag[i] - NpgMag[i]) * 10); }
      else
        { FminusNdm = NODATAm128; }
/*
   printf("(%s\", %ld, %ld, %d, %d, %d, %d, %d, %d, %d, %f, %d)",
  fileroot,RAcs,DECcs,Epoch00c,Fmm,Jmm,Vmm,Nmm,Cl,EXTcs,Ecc,PosAng);
*/

/* First check "sourceStatus" */
//      if (sourceStatus[i] < bad_flag && ! isnan(sourceStatus[i]))
      if (! isnan(sourceStatus[i]))
      {
        sprintf(ins_parvals,
        "(%ld, %ld, %ld, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %d, %ld);",
        gscID,RAcs,DECcs,Epoch00c,RaPMdmas,DecPMdmas,RaPMerr2mas,DecPMerr2mas,
        Fmm,Jmm,Vmm,Nmm,JminusFdm,FminusNdm,Cl,EXTcs,Ecc,PosAng,sourceStatus[i]);

//       strcpy(ins_comm,ins_parnames);
//       strcat(ins_comm,ins_parvals);
        sprintf(ins_comm, "INSERT INTO %s VALUES %s", db_tabname, ins_parvals);
/* printf("%s \n",ins_comm); */

        ret=db_query(0, ins_comm);
        if (!ret)
        {
          printf("DB error: %s\n", db_error(0));
          printf("     RA (rad)         Dec (rad)      RAcs     DECcs \n");
          printf("%16.12f %16.12f  %ld  %ld \n", RArad[i], Decrad[i], RAcs, DECcs);
          fprintf(dblent, "%ld %ld %ld %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %ld %s\n",
            gscID,RAcs,DECcs,Epoch00c,RaPMdmas,DecPMdmas,
            RaPMerr2mas,DecPMerr2mas,
            Fmm,Jmm,Vmm,Nmm,JminusFdm,FminusNdm,Cl,EXTcs,Ecc,PosAng,
            sourceStatus[i], filename);
        }

      } else {
        printf("sourceStatus %ld\n", sourceStatus[i]);
        fprintf(dblent, "%ld %ld %ld %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %ld %s\n",
          gscID,RAcs,DECcs,Epoch00c,RaPMdmas,DecPMdmas,
          RaPMerr2mas,DecPMerr2mas,
          Fmm,Jmm,Vmm,Nmm,JminusFdm,FminusNdm,Cl,EXTcs,Ecc,PosAng,
          sourceStatus[i], filename);
      }

      if (Epoch00c<=0)
      {
        printf("Bad Epoch: %f\n", meanEpoch[i]);
        printf("     RA (rad)         Dec (rad)      RAcs     DECcs \n");
        printf("%16.12f %16.12f  %ld  %ld \n", RArad[i], Decrad[i], RAcs, DECcs);
        printf("%ld %ld %ld %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %ld %s\n",
          gscID,RAcs,DECcs,Epoch00c,RaPMdmas,DecPMdmas,
          RaPMerr2mas,DecPMerr2mas,
          Fmm,Jmm,Vmm,Nmm,JminusFdm,FminusNdm,Cl,EXTcs,Ecc,PosAng,
          sourceStatus[i], filename);
        Epoch00c = 0;
      }

    }
/* Unlock the table for each read file */
    strcpy(ins_comm, "UNLOCK TABLES");
    ret=db_query(0, ins_comm);
    if (!ret) { printf("DB error: %s\n", db_error(0));  return(1); }

  }

  free (gscID2);
  free (RArad);
  free (Decrad);
  free (meanEpoch);
  free (RaPM);
  free (DecPM);
  free (RaPMerr);
  free (DecPMerr);
  free (FpgMag);
  free (JpgMag);
  free (VMag);
  free (NpgMag);
  free (Class);
  free (semiMajorAxis);
  free (eccentricity);
  free (PosAngle);
  free (sourceStatus);

  db_close(0);
  fclose( inlis );
  fclose( dblent );
//  fclose( oldepoch );
  return(0);

}
