@mcs_usrlib   ; User contributed Functions and Pro

PRO DBExample_reg, Host=host, User=user, Password=password, DBName=dbname, $
              Query=query, Debug=debug
;+
; NAME:
;       DBExample_reg
;
; PURPOSE:
;   MCS demonstration program using the routines in the "mcs_usrlib" library.
;
;   Use the HTM 6 "regions" table in mpe2018db DB as default.
;
;   Show usage of DBExec_Qry, TabStruct, TabFill, FieldInfo, DBFree and
;   transfer of a column of the output table into an array.
;   Here the output table is parsed into an array of structures (one
;   structure per row).
;
; CATEGORY:
;   MCS demonstration program using the mcs_usrlib library.
;
; CALLING SEQUENCE:
;   DBExample_reg [, see keywords section]
;
; INPUT PARAMETERS:
;   None.
;
; INPUT KEYWORDS:
;      Host: The DB server host name (def. localhost).
;      User: The user name to use for DB access.
;  Password: The password to use for DB access.
;    DBName: The database name to access.
;     Query: The complete query string.
;     Debug: If set then print some debug info.
;
; NOTES:
;   Must run from a directory where the command "idl2mcs" has been issued.
;   It create logical links to idl2mcs.dlm and idl2mcs.so.
;
;   To use the default query, the table "regions" must be present in the
;   selected database (can get it from ross.iasfbo.inaf.it/MCS/).
;
; COMMON BLOCKS:
;   Note that all the info are passed using COMMON blocks and it could be
;   useful to include "DBinfo" or/and "TABinfo" (see below).
;
;    TABinfo, mcs_types, is_rowstru, n_fields, n_rows, f_name, $
;             f_atype, f_itype, the_stru
;
;
; MODIFICATION HISTORY:
;   Written by L. Nicastro @ IASF-INAF, July 2006.
;  19/09/2008: LN, changed to use mcs_usrlib v0.3b.
;  03/12/2014: LN, changed for the IMPRS DB course.
;-

COMMON DBinfo
COMMON TABinfo

  in_qry = 0 & in_db = 0
; Keywords section (just check there is a query)
  IF (NOT KEYWORD_SET(query)) THEN BEGIN
    query = 'SELECT htmID_6,GSChtm FROM regions LIMIT 100'
    PRINT, 'Using query: ' + query
  ENDIF else $
    in_qry = 1

; DB section (here use MyCats as default DB)
  IF (NOT KEYWORD_SET(dbname)) THEN BEGIN
    dbname = 'mpe2018db'
    PRINT, 'Using DB: ' + dbname
  ENDIF else $
    in_db = 1

; Connect and execute the query using DBExec_Qry (return 0 if OK)
;PRINT, SYSTIME()
  status = DBExec_Qry(dbname, user, password, query, Host=host)
  IF (status EQ 0) THEN BEGIN
    TabStruct, Debug=debug ; Build the table structure 'tab_stru'
    TabFill                ; Fill in 'tab_stru'
  ENDIF
;PRINT, SYSTIME()

  PRINT, ""
  PRINT, "Fields info from FieldInfo:"

  FieldInfo, /INFO

  DBFree                   ; Free resources
; End DB section

  IF (status NE 0) THEN RETURN  ; Error?

  PRINT, 'Info on the returned structure (note the _NULL):'
  HELP, /STRU, tab_stru

  PRINT, ''
  PRINT, 'First 10 elements of the structure:'
  FOR i=0, MIN([9,n_rows-1]) DO BEGIN
    PRINT, i+1, " ", tab_stru[i]
  ENDFOR

  PRINT, ''
  PRINT, 'First 4 elements of each column:'
  FOR i=0, n_fields - 1 DO $ 
    PRINT, f_name[i], tab_stru[0:3].(i)

  PRINT, ''
  PRINT, 'Transfer first 10 elements of the fist column into a vector "c1":'
  icol = 0
  c1 = tab_stru[0:9].(icol)

  HELP, c1
  PRINT, c1[0:9]


; Let's use DBExec_ShowQry: this time we ask for a structure of arrays
; Note that we use common block saved (c_) info. They were set in the call
; to DBExec_Qry in case they where not given as input kwds

  IF ((NOT in_qry) AND (NOT in_db)) THEN BEGIN  ; only if query and db were not given
    query = 'SELECT * FROM regions limit 10'
    PRINT, 'Using query: ' + query

    PRINT, '... and returning data as an array of structures:'
    status = DBExec_ShowQry(c_dbname, c_user, c_password, query, Host=host, /ARR_STRU)

    IF (status NE 0) THEN RETURN  ; Error?

    PRINT, 'Info on the returned structure:'
    HELP, /STRU, tab_stru

  ENDIF

END
