@mcs_healplib   ; MCS-Base and MCS-HEALPix user contributed Functions and Pro

PRO sid_mcshpdemo1, Host=host, User=user, Password=password, DBName=dbname, $
    TABName=tabname, Order=order, Nested=nested, $
    OutFile=outfile, Units=units, Gnomic=gnomic, Debug=debug, $
; xxxView passed options
    PNG=png, _EXTRA=e
;+
; NAME:
;       MCSHPDEMO1
;
; PURPOSE:
;   MCS-SID-HEALPix demo program which shows how to obtain a map with the
;   number of occurences per pixel of a WHERE filtered query.
;
;   Show usage of DBExec_Qry, healp_MapFill, healp_MapSave.
;
; CATEGORY:
;   MCS-SID demonstration program using the mcs_usrlib/mcs_healplib libraries.
;
; CALLING SEQUENCE:
;   mcshpdemo1 [, see keywords section]
;
; INPUT PARAMETERS:
;   None.
;
; INPUT KEYWORDS:
;      Host: The DB server host name (def. localhost).
;      User: The user name to use for DB access (def. mpeusr).
;  Password: The password to use for DB access (def. mpe2018pass).
;    DBName: The database name to access (def. mpe2018db).
;   TABName: The reference table name (def. ucac2_initial).
;     Order: The output map Order (resolution level).
;    Nested: If set then assume NESTED scheme (used only if Order kwd passed).
;   OutFile: If set then save in that file (def. into "mcsdb_map.fits").
;     Units: Units of output data (used for the plot).
;    Gnomic: If set then plot Gnomic view (see gnomview help).
;     Debug: If set then print some debug info.
;       PNG: If set then save map into PNG files (see mollview help).
;
; Any extra keyword is passed to mollview / orthview / gnomview.
;
; NOTES:
;   To use the default query, the HEALPix indexed table "ucac2_initial" must be
;   present in the selected database (can get it from ross.iasfbo.inaf.it/MCS/).
;   Here we assume tables are SID indexed with HEALPix order=8,
;   schema=NESTED and column name is "healp8".
;
;   The HEALPix IDL library (see healpix.jpl.nasa.gov/) is required.
;
;   It is assumed that the query output just 1 column, i.e. the pixel IDs.
;   In case constant factors are used in the query, then it must be in
;   exponential notation (this is for MySQL - MCS communication reasons)!
;   For example: SELECT healpID, Mag1/1e2 FROM ....
;
; COMMON BLOCKS:
;   Note that all the info are passed using COMMON blocks and typically
;   in a main PRO need to include HEALPmap (see below).
;
;    COMMON DBptr, dbc, dbq
;    COMMON DBinfo,  c_host, c_user, c_password, c_dbname, c_query
;    HEALPmap, hp_nested, hp_nside, hp_order, hp_npix, $
;              hp_emptypix, hp_npixok, hp_map_nocc, hp_map_sum
;
;
; MODIFICATION HISTORY:
;   Written by L. Nicastro @ INAF-OAS, June 2018.
;  05/06/2018: LN: assume test catalogues were SID indexed with
;                  HEALPix order=8, schema=NESTED.
;-

; This is the common section for the "mcs_usrlib"
COMMON DBptr, dbc, dbq
COMMON DBinfo,  c_host, c_user, c_password, c_dbname, c_query
COMMON HEALPmap, hp_nested, hp_nside, hp_order, hp_npix, $
                 hp_emptypix, hp_npixok, hp_map_nocc, hp_map_sum

;Error handler
  CATCH, Error_status
  IF Error_status NE 0 THEN BEGIN
    PRINT, !ERROR_STATE.MSG
    DBClose
    RETURN
  ENDIF

; Keywords section
  order = 8
  hp_order = order

  IF (NOT KEYWORD_SET(tabname)) THEN $
    tabname = 'ucac2_initial'

  IF (NOT KEYWORD_SET(dbname)) THEN $
    dbname = 'mpe2018db'

  IF (NOT KEYWORD_SET(user)) THEN $
    user = 'mpeusr'

  IF (NOT KEYWORD_SET(password)) THEN $
    password = 'mpe2018pass'

  query1 = 'CALL SID.SelectCircleHEALP("myregion_circ", "*", "'+ dbname +'.'+ tabname +'", "healp8", 8, "RAmas/3.6e6", "DECmas/3.6e6", 0, 0, 1500, "")'

  PRINT, 'Query1: ', query1

  query2 = 'SELECT HEALPlookup(0,8, ramas/3.6e6, decmas/3.6e6) as healpID_nest_8 FROM SID.myregion_circ'

  PRINT, 'Query2: ', query2



; Open DB conn. and execute the first query
  DBConnect, dbname, user, password

  IF (dbc EQ ifd_null()) THEN $
    MESSAGE, 'DB connection is closed.'

  dbq1 = new_Query(ifd_null(), dbc, 0)
  Query_prepare, dbq1, query1
  Query_execute, dbq1, 1

  IF (ifd_got_error()) THEN BEGIN
    msg = ifd_last_error()
    ifd_reset_error
    del_Query, dbq1
    MESSAGE, msg
  ENDIF

; Free and destroy local query object
  Query_close, dbq1

  IF (ifd_got_error()) THEN BEGIN
    msg = ifd_last_error()
    ifd_reset_error
    MESSAGE, msg
  ENDIF
  del_Query, dbq1


; In the second query pass the parameters set in the first one
  err = DBExec_Qry(c_dbname, c_user, c_password, query2, Host=host)

; If order passed then can also pass the scheme
    IF (KEYWORD_SET(nested)) THEN $
      hp_nested = 1 $
    ELSE $
      hp_nested = 0

  IF (err EQ 0) THEN BEGIN
    err = healp_MapFill(/INFO)
    IF (hp_npixok EQ 0) THEN $
      err = 1  ; Error
  ENDIF
  DBClose                         ; Free resources

  IF (err NE 0) THEN RETURN  ; Error?

; Units and plot title
  IF (NOT KEYWORD_SET(units)) THEN BEGIN
     units = 'Nr of entries'
  ENDIF
  ptitle = 'MCS - SID demo plot: '+ tabname

; Some views: see mollview help (IDL> doc_library,'mollview')
;  mollview, hp_map_nocc, /online, graticule=[30,30], coord=['C','C'], colt=6, $
;    GIF=gif, PNG=png
;  orthview, hp_map_nocc, /online, graticule=[30,30], coord=['C','C'], colt=6, $
;    GIF=gif, PNG=png

  IF (KEYWORD_SET(gnomic)) THEN BEGIN
    gnomview, hp_map_nocc, /GRATICULE, COORD=['C','C'], $
              TITLE=ptitle, UNITS=units, NESTED=hp_nested, $
              PNG=png, _EXTRA=e
  ENDIF ELSE BEGIN
    mollview, hp_map_nocc, /GRATICULE, COORD=['C','C'], $
              TITLE=ptitle, UNITS=units, NESTED=hp_nested, $
              PNG=png, _EXTRA=e
    orthview, hp_map_nocc, /GRATICULE, COORD=['C','C'], ROT=[30,30], $
              TITLE=ptitle, UNITS=units, NESTED=hp_nested, $
              PNG=png, _EXTRA=e
  ENDELSE

PRINT, 'There are '+ STRTRIM(hp_npixok, 1) +' not empty pixels.'

; List of non empty pixels
;  idxs = where(hp_map_nocc gt 0,nfill)
;print, 'There are '+ strtrim(nfill,1) +' not empty pixels.'
;print, idxs
;print, hp_map_nocc[idxs]

  IF (KEYWORD_SET(outfile)) THEN BEGIN
    PRINT, 'Saving into file.'
    err = healp_MapSave(OUTFILE=outfile, UNITS=units)
  ENDIF

END
