#=
Efficiently load MySQL database tables from FITS files.

  Last changed: LN@INAF-OAS, 07/06/2018

Requirements:
Julia (v >= 0.6.0)
Julia packages: ArgParse, NamedTuples, MySQL, FITSIO

Add them with the following commands:
Pkg.add("ArgParse")
Pkg.add("NamedTuples")
Pkg.add("MySQL")
Pkg.clone("https://github.com/JuliaAstro/FITSIO.jl.git")
Pkg.build("FITSIO")

Note that the master version of the FITSIO package is required.

usage: fits2db.jl -u USER -p PASSWORD -d DATABASE [-s SERVER] [-H HDU]
                  [-v] [-h] FITS

positional arguments:
  FITS                  FITS file

optional arguments:
  -u, --user USER       User name
  -p, --password PASSWORD
                        Password
  -d, --database DATABASE
                        Database name
  -s, --server SERVER   Server host (default: "localhost")
  -H, --HDU HDU         HDU to import, starting from 1 (default: last one)
                        (type: Int64, default: -1)
  -v, --verbose         Be verbose
  -h, --help            Show this help message and exit



Example:
julia fits2db.jl -u root -p "root_password" -d Catalogs -v -H 2 skymapper.fits.gz
=#


using ArgParse, NamedTuples, MySQL, FITSIO

argparse = ArgParseSettings()
@add_arg_table argparse begin
    "--user", "-u"
    help = "User name"
    arg_type = String
    required = true
    "--password", "-p"
    help = "Password"
    arg_type = String
    required = true
    "--database", "-d"
    help = "Database name"
    arg_type = String
    required = true
    "--server", "-s"
    help = "Server host"
    arg_type = String
    default = "localhost"
    "--HDU", "-H"
    help = "HDU to import, starting from 1 (default: last one)"
    arg_type = Int
    default = -1
    "--verbose", "-v"
    help = "Be verbose"
    action = :store_true
    "FITS"
    help = "FITS file"
    required = true
end

args = parse_args(argparse)

println("Parsed args:")
for (arg,val) in args
    println("  $arg  =>  $val")
end

f = FITS(args["FITS"])
if args["HDU"] == -1
    args["HDU"] = length(f)
end

hdu = f[args["HDU"]]
show(hdu)
println()
header = read_header(hdu);
nrows = header["NAXIS2"]

data = []
fieldNames = FITSIO.colnames(hdu)
for fieldName in fieldNames
    col = read(hdu, fieldName)
    push!(data, col)
end

ncols = length(fieldNames)

create = Vector{String}(0)
for i in 1:ncols
    cell = data[i][1]
    sqlType = ""
    if isa(cell, Bool)
        sqlType = "CHAR(1)"
    elseif isa(cell, Int8)
        sqlType = "TINYINT"
    elseif isa(cell, Int16)
        sqlType = "SMALLINT"
    elseif isa(cell, Int32)
        sqlType = "INT"
    elseif isa(cell, Int64)
        sqlType = "BIGINT"
    elseif isa(cell, Float32)
        sqlType = "FLOAT"
    elseif isa(cell, Float64)
        sqlType = "DOUBLE"
    elseif isa(cell, String)
        lf = chop(header["TFORM"*string(i)])
        sqlType = "VARCHAR($lf)"
    else
        error("Unsupported type: " * string(typeof(cell)))
    end
    push!(create, "`"* fieldNames[i] * "` " * sqlType)
end


tableName = basename(args["FITS"])
tableName = (split(tableName, "."))[1] * "_" * string(args["HDU"])

create = "CREATE TABLE `$tableName` (" * join(create, ", ") * ")"
println(create)

const conn = MySQL.connect(args["server"], args["user"], args["password"])
MySQL.execute!(conn, "use " * args["database"]) 
MySQL.execute!(conn, "DROP TABLE IF EXISTS `$tableName`")
MySQL.execute!(conn, create)

cacheSize = Int(floor(65000. / ncols))
if cacheSize < 1
    error("Too many fields: $ncols")
end

placeholder = Array{String}(ncols)
placeholder[:] = "?"
placeholder = "(" * join(placeholder, ",") * ")"
all = placeholder
for i in 2:cacheSize
    all *= ", $placeholder"
end
query = "INSERT INTO $tableName VALUES $all"
#println(query)
stmt = MySQL.Stmt(conn, query)
affrows = 0

values = []
irow = 1
while irow <= nrows
    toBeWritten = nrows - irow
    if (toBeWritten < cacheSize)
        cacheSize = toBeWritten + 1;
        placeholder = Array{String}(ncols)
        placeholder[:] = "?"
        placeholder = "(" * join(placeholder, ",") * ")"
        all = placeholder
        for i in 2:cacheSize
            all *= ", $placeholder"
        end
        query = "INSERT INTO $tableName VALUES $all"
        #println(query)
        stmt = MySQL.Stmt(conn, query)
    end

    values = []
    for iCache in 1:cacheSize
        for icol in 1:ncols
            cell = data[icol][irow]
            if !isa(cell, String)  &&  ~isfinite(cell)
                cell = nothing
            end
            # println(fieldNames[icol], " ", cell)
            push!(values, cell)
        end
        irow += 1
    end
    #println(length(values))
    affrows += MySQL.execute!(stmt, values)        
end

close(f)

println("$affrows record(s) written")
