/*
   myCat_qry.cpp

  Purpose:

   Fortran callable routine to build the query string to read custom
   MySQL DBs.

   It only requires the RA, DEC corners (in deg.) of a rectangle,
   or the center of a square ("RAdeg DECdeg" or "hh mm ss dd mm ss") and
   its side in arcmin.
   If the flag "is_circle" is not 0, then assumes a circle and the third
   parameter is the radius.

   The resulting MySQL query returns all the data for the objects found in that
   rectangle/square/circle. No sorting is performed.

  23/06/2010: DIF 0.5.2 compatibility.
  06/12/2014: removed offset (32768) for HTM 6 depth index.

   LN@IASF-INAF, Feb. 2007                   ( Last change: 06/12/2014 )
*/

using namespace std;

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <VarVecDef.h>
#include "SpatialVector.h"
#include "SpatialInterface.h"
#include "SpatialDomain.h"

// Offset to reset index in 0 - 32767: 32768 regions in the order 6 HTM grid
//#define OFFSET 32768


int wcount(const char *st) {
  int l=0, nw=0, len;

  len = strlen(st);
  while (len > 0) {
    if (st[len-1] == ' ') {
      if (l>0) {
        ++nw;
        l = 0;
      }
    } else ++l;
    --len;
  }
  if (l > 0) ++nw;

  return nw;
}


char *trim (char *str)
{
    char *ibuf, *obuf;

    if (str)
    {
          for (ibuf = obuf = str; *ibuf; )
          {
                while (*ibuf && (*ibuf==' '))
                      ibuf++;
                if (*ibuf && (obuf != str))
                      *(obuf++) = ' ';
                while (*ibuf && (*ibuf!=' '))
                      *(obuf++) = *(ibuf++);
          }
          *obuf = 0;
    }
    return (str);
}

// Trim both leading and trailing spaces
string trimS(const string& str)
{
// Find the first non-blank
  size_t start = str.find_first_not_of(" \t\n\r");
// If all spaces or empty return an empty string
  if (start == string::npos) return "";

  return str.substr(start, str.find_last_not_of(" \t\n\r") - start + 1 );
}


double deg_ra(const char *ra_str)
{
  unsigned int rah,ram;
  float rafs;
  double ra;
//  static char str_ra[20];

  sscanf(ra_str,"%2d", &rah);
  ra_str += 3;
  sscanf(ra_str,"%2d", &ram);
  ra_str += 3;
  sscanf(ra_str,"%f", &rafs);
  ra_str -= 6;
  ra  = (rah + ram/60. + rafs/3.6e3) * 15.;
  if (ra < 0.) ra *= -1;
//  printf("ra=%f\n",ra);
  return (ra);
}

double deg_dec(const char *dec_str)
{
  unsigned int deg, dem;
  float defs;
  double de;
  char sign[2];

//  sscanf(dec_str,"%1s%2d%1s%2d%1s%f",&sign, &deg, &div, &dem, &div, &defs);
  *sign = dec_str[0];
  if (*sign != '+' && *sign != '-') {
    sscanf(dec_str,"%2d", &deg);
    dec_str += 3;
  } else {
    sscanf(dec_str,"%1s%2d",sign, &deg);
    dec_str += 4;
  }
  sscanf(dec_str,"%2d", &dem);
  dec_str += 3;
  sscanf(dec_str,"%f", &defs);
  dec_str -= 7;
  de  = deg + dem/60. + defs/3.6e3;
  if (sign[0] == '-') de *= -1;
//  printf("dec=%f\n",de);
  return (de);
}


/*
   getIntersect6

  Purpose:

   Intersect a rectangular domain with the HTM depth 6 grid returning the IDs
   of full and partial nodes

  Input RA, DEC (in deg.) of the corners.

  Notes:
   Return original IDs rather than in the range 0 - 32767

   LN@IASF-CNR, January 2003                   ( Last change: 06/12/2014 )
*/

int getIntersect6(float64 *ra, float64 *dec,
                  ValVec<uint64> *flist2, ValVec<uint64> *plist2)
{
  int depth=6;
  ValVec<uint64> plist, flist;	// List results
  size_t i;

  flist2->cut(flist2->length());
  plist2->cut(plist2->length());

  try
  {
// construct index with given depth and savedepth
//     htmInterface htm(depth,savedepth);  // generate htm interface
//     const SpatialIndex &index = htm.index();
    const SpatialIndex index(depth);
    SpatialDomain domain;    // initialize empty domain


    SpatialVector v1(ra[0], dec[0]), v2(ra[1], dec[1]),
                  v3(ra[2], dec[2]), v4(ra[3], dec[3]);
    SpatialConvex cvx(&v1,&v2,&v3,&v4);
    domain.add(cvx);

    domain.intersect(&index,plist,flist);     // intersect with list
/*
    for(i = 0; i < flist.length(); i++){      // just loop full nodes list
      flist(i) -= OFFSET;                     // subtracting OFFSET
    }
    for(i = 0; i < plist.length(); i++){      // just loop partial nodes list
      plist(i) -= OFFSET;                     // subtracting OFFSET
    }
*/
  }
  catch (SpatialException &x)
  {
    return -1;
  }

  *flist2 = flist;
  *plist2 = plist;
  return 0;
}


/*
   getcIntersect6

  Purpose:

   Intersect a circular domain with the HTM depth 6 grid returning the full
   and partial nodes.

  Input RA, DEC and radius (in deg.) of the circle.

  Notes:
   Return original IDs rather than in the range 0 - 32767

   LN@IASF-INAF, Dec. 2006                   ( Last change: 06/12/2014 )
*/

int getcIntersect6(float64 ra, float64 dec, float64 radius,
                   ValVec<uint64> *flist2, ValVec<uint64> *plist2)
{
/* degrees to radians */
  static const double DEG2RAD = 1.74532925199432957692369E-2;
  int depth=6;
  ValVec<uint64> plist, flist;	// List results
  size_t i;
 
  float64 distance = cos(radius*DEG2RAD);

  flist2->cut(flist2->length());
  plist2->cut(plist2->length());

  try
  {
    const SpatialIndex index(depth);

    SpatialDomain domain;    // initialize empty domain
    SpatialVector v(ra, dec);
    SpatialConvex cvx;
    SpatialConstraint constr(v, distance);
    cvx.add(constr);
    domain.add(cvx);


    domain.intersect(&index,plist,flist);     // intersect with list
/*
    for(i = 0; i < flist.length(); i++){      // just loop full nodes list
      flist(i) -= OFFSET;                     // subtracting OFFSET
    }
    for(i = 0; i < plist.length(); i++){      // just loop partial nodes list
      plist(i) -= OFFSET;                     // subtracting OFFSET
    }
*/
  }
  catch (SpatialException &x)
  {
    return -1;
  }

  *flist2 = flist;
  *plist2 = plist;
  return 0;
}


// Below to convert an integer/long to a string.
string itos(int i)
{
  char buf[20];
  sprintf(buf, "%d", i);
  return string(buf);
}

string ltos(long i)
{
  char buf[20];
  sprintf(buf, "%ld", i);
  return string(buf);
}


/*
   Default number of required arguments is 4 -> rectangle corners.
   Could be 3 for a square (center + side size) or circle.
*/

extern "C" {

void
mycat_qry_(char *coords, int *coords_len, char *catName, int *catName_len,
    int *is_circle,
    char *qry, int *qry_len) {

static short NVALS = 4;

  int nw;                     // N words in coords string
  unsigned short is_neg;      // Check for negative Dec (like -00 ...)
  unsigned short is_square=0; // Toggle for input square center and size
  unsigned short ra0border=0; // Toggle for negative start RA range
  unsigned int i;
  int j, nnodes;
  float64 ra[NVALS], dec[NVALS], side_arcm=0., hside_deg=0.,
          rac=0., decc=0., ra0, ra2, dec0, dec2;
  unsigned long  ramin,ramax;
  long  decmin,decmax;
  unsigned int ibl, lbl;
  string cc, ra_str, de_str;
  ValVec<uint64> plist, flist;

  static const double DEG2RAD=1.745329251994329577E-2,
      DEG2CAS=3.6E5;  // DEG2MAS=3.6E6

  string id_strn = "htmID_6=";
  string racs_strn = "RAcs";     // Change to appropriate name
  string deccs_strn = "DECcs";   //      "   "
  string str_htmid="", str_coord="", comm_str="";

// do not order       ") ORDER BY RAcs ASC"};
//  char id_strv[LIDV], racs_strv[LRAV], deccs_strv[LDECV];  //, dec_deg[4];

  string scoords,CATName;
  scoords.assign(coords,*coords_len-1);
  CATName.assign(catName,*catName_len-1);
  CATName = trimS(CATName);

  nw = wcount(scoords.c_str());
/* There must be at least 3 args. */
  if (nw != NVALS)
  {
    if (nw >= 7)  // RA (hh mm ss) Dec. (dd mm ss) Rad
    {
      ibl = 0;
      lbl = scoords.find(" ",ibl);
      cc = scoords.substr(ibl,lbl);
      rac = atof(cc.c_str());
      ibl = lbl+1;
      lbl = scoords.find(" ",ibl);
      cc = scoords.substr(ibl,lbl-ibl);
      rac += atof(cc.c_str())/60.;
      ibl = lbl+1;
      lbl = scoords.find(" ",ibl);
      cc = scoords.substr(ibl,lbl-ibl);
      rac += atof(cc.c_str())/3600.;
      rac *= 15.;
/* Cannot use strncmp as the string could start with blanks */
      ibl = lbl+1;
      lbl = scoords.find(" ",ibl);
      cc = scoords.substr(ibl,lbl-ibl);

//      if ( strchr(dec_deg,'-') != NULL )
      if ( cc[0] == '-')
      {
        is_neg = 1;
        decc = - atof(cc.c_str());
      } else {
        is_neg = 0;
        decc = atof(cc.c_str());
      }
      ibl = lbl+1;
      lbl = scoords.find(" ",ibl);
      cc = scoords.substr(ibl,lbl-ibl);
      decc += atof(cc.c_str())/60.;
      ibl = lbl+1;
      lbl = scoords.find(" ",ibl);
      cc = scoords.substr(ibl,lbl-ibl);
      decc += atof(cc.c_str())/3600.;

      if ( is_neg ) decc *= -1;
    } else if (nw != 3)
    {
       *qry = 0;
      return;
    } else {  // square or circle
/* Check if RA and Dec are given as string with ":" */
      lbl = scoords.find(" ",0);
      cc = scoords.substr(0,lbl);
      if (scoords.find(":",0) != string::npos)
      {
        rac  = deg_ra(cc.c_str()); 
        ibl = lbl+1;
        lbl = scoords.find(" ",ibl);
        cc = scoords.substr(ibl,lbl-ibl);
//        if ( cc[0] != '+' && cc[0] != '-') cc = "+" + cc;
        decc = deg_dec(cc.c_str()); 
      } else {
        ra_str = cc;
        rac  = atof(cc.c_str());
        ibl = lbl+1;
        lbl = scoords.find(" ",ibl);
        cc = scoords.substr(ibl,lbl-ibl);
        de_str = cc;
        decc = atof(cc.c_str());
      }
    }
    is_square = 1;
  }

  if (is_square)  // but could also be circle
  {
    cc = scoords.substr(lbl);
    side_arcm = atof(cc.c_str());
//printf("rac,decc,side_arcm %f %f %f\n",rac,decc,side_arcm);
    if (side_arcm <= 0.)
    {
      cout << "You gave a negative side: " << side_arcm << endl;;
      *qry = 0;
      return;
    }
    if (*is_circle)
      hside_deg = side_arcm/60.;
    else
      hside_deg = side_arcm/2./60.;
     ra[0] =  rac - hside_deg/cos(decc*DEG2RAD);
    dec[0] = decc - hside_deg;
     ra[2] =  rac + hside_deg/cos(decc*DEG2RAD);
    dec[2] = decc + hside_deg;
  } else {
      lbl = scoords.find(" ",0);
      cc = scoords.substr(0,lbl);
     ra[0] = atof(cc.c_str());
     ibl = lbl+1;
     lbl = scoords.find(" ",ibl);
     cc = scoords.substr(ibl,lbl-ibl);
    dec[0] = atof(cc.c_str());
     ibl = lbl+1;
     lbl = scoords.find(" ",ibl);
     cc = scoords.substr(ibl,lbl-ibl);
     ra[2] = atof(cc.c_str());
     ibl = lbl+1;
     lbl = scoords.find(" ",ibl);
     cc = scoords.substr(ibl,lbl-ibl);
    dec[2] = atof(cc.c_str());
    rac  = (ra[0]+ra[2])/2.;
    decc = (dec[0]+dec[2])/2.;
  }

/* Order ranges */
  ra0 = (ra[0] < ra[2]) ? ra[0] : ra[0] - 360.;
  ra2 = (ra[2] > ra0) ? ra[2] : ra0;
  dec0 = (dec[0] < dec[2]) ? dec[0] : dec[2];
  dec2 = (dec[2] > dec[0]) ? dec[2] : dec[0];

  if (ra0 == ra2 || dec0 == dec2)
  {
    cout << "Cannot give identical RA or DEC: " << scoords << endl;;
    *qry = 0;
    return;
  }

//  float64 skyarea;
//  if (*is_circle)
//    skyarea = M_PI * hside_deg*hside_deg;
//  else
//    skyarea = (ra2-ra0)*cos(decc*DEG2RAD) * (dec2-dec0);

/* Check for border effects: not all cases are considered here! */
  if (ra2 <= 0.)
  {
    cout << "Negative RA: " << ra0 <<" "<< ra2 << endl;
    *qry = 0;
    return;
  }
  if (ra0 < 0.)
  {
    ra0 += 360.;
    ra0border = 1;
  }

  if (dec0 >= 90. || dec2 <= -90.)
  {
    cout << "DEC out of range: " << dec0 <<" "<< dec2 << endl;
    *qry = 0;
    return;
  }
  dec0 = (dec0 < -90.) ? -90. : dec0;
  dec2 = (dec2 >  90.) ?  90. : dec2;

   ra[0] = ra[1] = ra0;
   ra[2] = ra[3] = ra2;
  dec[0] = dec[3] = dec0;
  dec[2] = dec[1] = dec2;

  if (*is_circle)
    getcIntersect6(rac, decc, hside_deg, &flist, &plist);
  else
    getIntersect6(ra, dec, &flist, &plist);

  nnodes = flist.length() + plist.length();
  if ( nnodes < 1 )
  {
    cout << "Big troubles: no region selected!" << endl;
    exit(1);
  }

/* Define integer ranges */
  ramin = (long)(ra0*DEG2CAS+0.5);
  ramax = (long)(ra2*DEG2CAS+0.5);
  decmin = (long)(dec0*DEG2CAS+0.5);
  decmax = (long)(dec2*DEG2CAS+0.5);

//printf("ramin,ramax,decmin,decmax: %lu %lu %ld %ld \n",ramin,ramax,decmin,decmax);

/* Allocate space for the "SELECT ..." etc. */

/* Build the htmIDs string: allocate space for the number of regions + " OR " */

/* Allocate space for the coordinates */

// printf("No. char. allocated for query string: %d \n",nct);

/* Simple case: only 1 region interested: MUST be partial */
  if (nnodes == 1 && plist.length() == 0)
  {
    cout << "Big 2 troubles: partial region expected!" << endl;
    exit(2);
  }

/* For full regions do not need to select in RA, DEC. */
  if (flist.length() > 0)
  {
    for (i=0; i<flist.length(); i++)
      str_htmid = str_htmid + id_strn + itos((unsigned int)(flist(i))) + " OR ";
    str_htmid += "(";
  }

  if (plist.length() > 0)
  {
    str_htmid += "(";

    for (i=0; i<plist.length(); i++)
    {
      if ( i > 0 )
        str_htmid += " OR ";
      str_htmid = str_htmid + id_strn + itos((unsigned int)(plist(i)));
    }
    str_htmid  += ") AND ";
  }

/* Build the coordinates range string */
  str_coord = "(" + racs_strn + ">" + ltos(ramin);

  if ( ra0border )
    str_coord += " OR ";
  else
    str_coord += " AND ";

  str_coord = str_coord + racs_strn + "<" + ltos(ramax) +
              ") AND (" + deccs_strn + ">" + ltos(decmin) +
              " AND " + deccs_strn + "<" + ltos(decmax) +
              ")";

  if (flist.length() > 0)
    str_coord += " )";

// Is circle?
  if (*is_circle) {
    string sdist = "sdist_cs(" + racs_strn + "," + deccs_strn + "," +
            ra_str + "," + de_str + ") AS d";
    comm_str = "SELECT *," + sdist + " FROM " + CATName + " WHERE(" +
             str_htmid + str_coord + ")" +
             " HAVING d<" + trimS(cc);
  } else

/* Build query string */
  comm_str = "SELECT * FROM " + CATName + " WHERE(" +
             str_htmid + str_coord + ")";


  strcpy(qry,comm_str.c_str());
// Null filling
  for (j=strlen(qry); j<*qry_len; j++)
    qry[j]='\0';
//cout << qry << "*"<<endl;
    return;

}

}
